library ieee;
use ieee.std_logic_1164.all;

entity moore_state_machine is
	port(
		clk	      : in  std_logic;
		reset_n     : in  std_logic;
		m           : in  std_logic_vector(2 downto 0);
		cnt         : in  std_logic_vector(4 downto 0);
		cnt_en      : out std_logic;
		cnt_reset   : out std_logic;
		perm_sel    : out std_logic;
		perm_wren   : out std_logic;
		perm_addr   : out std_logic_vector(1 downto 0);
		mux_sel     : out std_logic;
		reg_en      : out std_logic;
		ram_wren_a  : out std_logic;
		ram_wren_b  : out std_logic;
		cpu_en      : out std_logic;
		step        : out std_logic_vector(2 downto 0)
	);
end moore_state_machine;

architecture rtl of moore_state_machine is

  type state_type is (s0, s1, s2, s3, s4, s5, s6, s7, s8, s9, s10, s11, s12, s13);
  signal state : state_type;

begin

	process (clk, reset_n)
	begin
		if reset_n = '0' then
			state <= s0;
		elsif rising_edge(clk) then
			case state is
				when s0=>
					if m = "111" then
						state <= s0;
					else
						state <= s1;
					end if;
				when s1=>
					state <= s2;
				when s2=>
					state <= s3;
				when s3 =>
					state <= s4;
				when s4 =>
					state <= s5;
				when s5 =>
					state <= s6;
				when s6 =>
					state <= s7;
				when s7 =>
					state <= s8;
				when s8=>
					if m = "111" then
						state <= s9;
					else
						state <= s1;
					end if;
				when s9=>
					if m = "111" then
						state <= s10;
					else
						state <= s1;
					end if;
				when s10=>
					if m = "111" then
						state <= s11;
					else
						state <= s1;
					end if;
				when s11=>
					if m = "111" then
						state <= s12;
					else
						state <= s1;
					end if;
				when s12=>
					if m = "111" then
						state <= s13;
					else
						state <= s1;
					end if;
				when s13=>
				   if cnt = "10011" then
					   state <= s0;
					elsif m = "111" then
						state <= s8;
					else
						state <= s1;
					end if;
			end case;
		end if;
	end process;

	process (state)
	begin
		case state is
			when s0 =>
			  cnt_en <= '0';
		     cnt_reset <= '0';
		     perm_sel <= '0';
			  perm_wren <= '0';
			  perm_addr <= "00";
			  mux_sel <= '0';
	        reg_en <= '0';
			  ram_wren_a <= '0';
			  ram_wren_b <= '0';
			  cpu_en <= '1';
           step <= "000";
			when s1 =>
			  cnt_en <= '0';
		     cnt_reset <= '0';
		     perm_sel <= '0';
			  perm_wren <= '0';
			  perm_addr <= "00";
			  mux_sel <= '0';
	        reg_en <= '1';
			  ram_wren_a <= '0';
			  ram_wren_b <= '0';
			  cpu_en <= '0';
           step <= "001";
			when s2 =>
			  cnt_en <= '0';
		     cnt_reset <= '0';
		     perm_sel <= '0';
			  perm_wren <= '0';
			  perm_addr <= "00";
			  mux_sel <= '0';
	        reg_en <= '0';
			  ram_wren_a <= '1';
			  ram_wren_b <= '1';
			  cpu_en <= '0';
           step <= "010";
			when s3 =>
			  cnt_en <= '0';
		     cnt_reset <= '0';
		     perm_sel <= '0';
			  perm_wren <= '0';
			  perm_addr <= "00";
			  mux_sel <= '0';
	        reg_en <= '0';
			  ram_wren_a <= '0';
			  ram_wren_b <= '0';
			  cpu_en <= '0';
           step <= "011";
			when s4 =>
			  cnt_en <= '0';
		     cnt_reset <= '0';
		     perm_sel <= '0';
			  perm_wren <= '0';
			  perm_addr <= "00";
			  mux_sel <= '0';
	        reg_en <= '0';
			  ram_wren_a <= '1';
			  ram_wren_b <= '1';
			  cpu_en <= '0';
           step <= "100";
			when s5 =>
			  cnt_en <= '0';
		     cnt_reset <= '0';
		     perm_sel <= '0';
			  perm_wren <= '0';
			  perm_addr <= "00";
			  mux_sel <= '0';
	        reg_en <= '0';
			  ram_wren_a <= '0';
			  ram_wren_b <= '0';
			  cpu_en <= '0';
           step <= "101";
			when s6 =>
			  cnt_en <= '0';
		     cnt_reset <= '0';
		     perm_sel <= '0';
			  perm_wren <= '0';
			  perm_addr <= "00";
			  mux_sel <= '0';
	        reg_en <= '0';
			  ram_wren_a <= '1';
			  ram_wren_b <= '1';
			  cpu_en <= '0';
           step <= "110";
			when s7 =>
			  cnt_en <= '0';
		     cnt_reset <= '1';
		     perm_sel <= '0';
			  perm_wren <= '0';
			  perm_addr <= "00";
			  mux_sel <= '1';
	        reg_en <= '0';
			  ram_wren_a <= '1';
			  ram_wren_b <= '1';
			  cpu_en <= '0';
           step <= "111";
			when s8 =>
			  cnt_en <= '1';
		     cnt_reset <= '0';
		     perm_sel <= '1';
			  perm_wren <= '0';
			  perm_addr <= "00";
			  mux_sel <= '0';
	        reg_en <= '0';
			  ram_wren_a <= '0';
			  ram_wren_b <= '0';
			  cpu_en <= '1';
           step <= "000";
			when s9 =>
			  cnt_en <= '1';
		     cnt_reset <= '0';
		     perm_sel <= '1';
			  perm_wren <= '0';
			  perm_addr <= "00";
			  mux_sel <= '0';
	        reg_en <= '0';
			  ram_wren_a <= '0';
			  ram_wren_b <= '0';
			  cpu_en <= '1';
           step <= "000";
			when s10 =>
			  cnt_en <= '1';
		     cnt_reset <= '0';
		     perm_sel <= '1';
			  perm_wren <= '0';
			  perm_addr <= "01";
			  mux_sel <= '0';
	        reg_en <= '0';
			  ram_wren_a <= '0';
			  ram_wren_b <= '0';
			  cpu_en <= '1';
           step <= "000";
			when s11 =>
			  cnt_en <= '0';
		     cnt_reset <= '0';
		     perm_sel <= '1';
			  perm_wren <= '0';
			  perm_addr <= "10";
			  mux_sel <= '0';
	        reg_en <= '0';
			  ram_wren_a <= '0';
			  ram_wren_b <= '0';
			  cpu_en <= '1';
           step <= "000";
			when s12 =>
			  cnt_en <= '0';
		     cnt_reset <= '0';
		     perm_sel <= '1';
			  perm_wren <= '0';
			  perm_addr <= "11";
			  mux_sel <= '0';
	        reg_en <= '0';
			  ram_wren_a <= '0';
			  ram_wren_b <= '0';
			  cpu_en <= '1';
           step <= "000";
			when s13 =>
			  cnt_en <= '1';
		     cnt_reset <= '0';
		     perm_sel <= '1';
			  perm_wren <= '1';
			  perm_addr <= "00";
			  mux_sel <= '0';
	        reg_en <= '0';
			  ram_wren_a <= '1';
			  ram_wren_b <= '0';
			  cpu_en <= '1';
           step <= "000";
		end case;
	end process;

end rtl;
